from .lab_iat_schema import LAB_IAT_SCHEMA
import copy

LAB_IAT_RECTANGLE_SCHEMA = copy.deepcopy(LAB_IAT_SCHEMA)

# Enforce number of vertices
LAB_IAT_RECTANGLE_SCHEMA["properties"]["annotations"]["items"]["properties"]["polygons"]["items"]["properties"]["geometry"]["properties"]["vertices"] =  { "$ref" : "iat#/$defs/rectangle_vertices" }

# Enforce geometry type
LAB_IAT_RECTANGLE_SCHEMA["properties"]["annotations"]["items"]["properties"]["polygons"]["items"]["properties"]["geometry"]["properties"]["type"] = { "const" : "rectangle" }


"""Methods to handle labels formatting

KILI example

"categories": [
    {
        "name": "PARENT"
    }
],
"children": {
    "CLASSIFICATION_JOB": {
        "categories": [
            {
                "confidence": 100,
                "name": "ENFANT_1"
            }
        ]
    },
    "CLASSIFICATION_JOB_0": {
        "categories": [
            {
                "children": {
                    "TRANSCRIPTION_JOB_0": {
                        "text": "test 2"
                    }
                },
                "confidence": 100,
                "name": "ENFANT_2"
            },
            {
                "children": {
                    "TRANSCRIPTION_JOB": {
                        "text": "test"
                    }
                },
                "confidence": 100,
                "name": "ENFANT_1"
            }
        ]
    }
},

"""

from copy import deepcopy
from typing import Dict, List, Optional


def clean_up(labels : Dict):
    cleaned = deepcopy(labels)
    for list_id in cleaned:
        for label in cleaned[list_id]["labels"]:
            del label["uid"]
        if cleaned[list_id].get("children", None):
            cleaned[list_id]["children"] = clean_up(cleaned[list_id]["children"])
    return cleaned

def extract_main_labels(labels : Dict) -> List[Dict] :
    main_labels = []
    for list in labels:
        value = labels[list]
        for label in value['labels']:
            main_labels.append({
                "id" : label["id"],
                "name" : label["id"],
                "list" : list
            })
    
    return main_labels

def find_label_in_main_list(label_id : str, project_configuration : Dict) -> List[Dict] :
    mainLists = project_configuration["metadata"]["labelOptions"]["mainLists"]
    availableLabels = project_configuration["metadata"]["labelOptions"]["availableLabels"]
    for mlist in mainLists:
        mainListKey = mlist["name"]
        for datalistKey in mlist["keys"]:
            for label in availableLabels.get(datalistKey, []):
                if label["id"] == label_id:
                    return {
                        mainListKey : {
                            "labels" : [
                                {
                                    "id": label["id"],
                                    "name": label["name"]
                                }
                            ]
                        }
                    }
    return None

def flatten_available_labels_used_in_main_lists(project_configuration : Dict, label_attr : str = 'id') -> List[Dict] :
    mainLists = project_configuration["metadata"]["labelOptions"]["mainLists"]
    availableLabels = project_configuration["metadata"]["labelOptions"]["availableLabels"]
    labels_list = set()
    for mlist in mainLists:
        for datalistKey in mlist["keys"]:
            for label in availableLabels.get(datalistKey, []):
                labels_list.add(label[label_attr])
                
    return list(labels_list)

def extract_labels_map_by_id(project_configuration : Dict, field: Optional[str] = None) -> List[Dict] :
    availableLabels = project_configuration.get("metadata",{}).get("labelOptions", {}).get("availableLabels",{})

    mapping = {}

    for listId in availableLabels:
        for label in availableLabels[listId]:
            if field:
                mapping[label['id']] = label[field]
            else:
                mapping[label['id']] = label
    
    return mapping


from typing import BinaryIO, Dict, List

from isahitlab.api.base import BaseApi

from ..helpers import get_response_json, log_raise_for_status


class DatasetApi(BaseApi):
    """Dataset API Calls"""
    

    def upload_file(self, project_id: str, dataset_id: str, file : BinaryIO, folder: str) -> Dict :
        """Upload file to dataset"""
        
        files = {'file': file }

        uploaded = self._http_client.post('api/file-manager/datasets/{}/resources'.format(dataset_id), files=files, params={ "projectId" : project_id }, data={ "path": folder })
        
        log_raise_for_status(uploaded)
        
        return get_response_json(uploaded)


from abc import ABC
from isahitlab.core.http.http_client import HttpClient


class BaseAction(ABC):
    
    http_client: HttpClient # instantianted in the isahitlab client child class

    def __init__(self, http_client: HttpClient) -> None:
        """Initialize Base Operation Class."""
        self._http_client = http_client


from dataclasses import dataclass
from typing import List, Literal, NewType, Optional

from isahitlab.domain.project import ProjectId

from .pagination import PaginationFilters

BatchId = NewType("BatchId", str)
BatchStatus = Literal["pending", "complete", "configuring"]

@dataclass
class BatchFilters(PaginationFilters):
    """Batch filters for running a batch search."""

    project_id: Optional[ProjectId] = None
    status_in: Optional[List[BatchStatus]] = None
    search: Optional[str] = None

@dataclass
class BatchPayload:
    """Batch payload
    
    Args:
        name: Name of the batch
    """

    name: str



from typing import NewType, Union, Tuple
import io
from dataclasses import dataclass

DatasetId = NewType("DatasetId", str)

@dataclass
class FilePayload(dict):
    """Task payload for task create

    Inherit from dict to make it JSON serializable
    
    Args:
        file: Name of the task (appended to the path, it must be unique in a dataset)
        path: Path of a resource (ex.: "./images/image1.jpg")
    """

    file: Union[str, Tuple[str, io.IOBase]]
    path:  Union[str, None]
    

    def __init__(self, file, path):
        self.file = file
        self.path = path
        dict.__init__(self, file=file, path=path)



from typing import Generator, Dict, List, Union, Optional, Any
import time
from isahitlab.api.base import BaseApi
from isahitlab.domain.task import TaskFilters
from isahitlab.domain.batch import BatchId
from isahitlab.domain.project import ProjectId
from isahitlab.helpers.list import divide_chunks
from .mappers import task_filters_mapper
from ..helpers import get_response_json, log_raise_for_status, Paginated

class TaskApi(BaseApi):
    """Task API Calls"""

    def get_all_tasks(self, filters: TaskFilters) -> Generator[tuple[Dict, int, int], None, None] :
        """Get tasks"""

        page = 1
        limit = 100

        filters.pagination = True
        filters.limit = limit

        while True:
            filters.page = page

            paginated: Paginated = self._get_tasks_page(filters)
            docs = paginated['docs']
            page = paginated['nextPage']

            # Count already loaded
            totalDocs = paginated['totalDocs']
            loaded = (paginated['page'] - 1) * paginated['limit'] + len(docs)

            yield (docs, loaded, totalDocs)

            if not paginated['hasNextPage']:
                break

            time.sleep(0.5)

    def get_existing_task_names(self, batch_id : BatchId, names: List[str]) -> Dict[str, List[str]]:
        """Get existing task names by chunk"""
        chunks = divide_chunks(names, 500)

        duplicates = []

        for chunk in chunks:
            res = self._get_existing_task_names_chunk(batch_id, chunk)
            if len(res["duplicate"]) > 0:
                duplicates = duplicates + res["duplicate"]

        return duplicates


    def create_task(self, batch_id : BatchId, name : str, resources : List[Union[str, Dict]], data_id : Union[Dict, None], properties: Optional[Dict[str,Any]] = None, grant_resources: Optional[str] = None, grant_data: Optional[str] = None) -> Dict:
        """Create task"""
        
        data = { 
            "name": name,
            "resources" : resources,
            "properties" : properties,
            "dataId" : data_id,
            "grantResources": grant_resources,
            "grantData": grant_data,
        }
        
        created = self._http_client.post('api/task-manager/batches/{}/tasks'.format(batch_id), json=data)

        log_raise_for_status(created)

        return get_response_json(created)
    
    def update_properties(self, project_id : ProjectId, task_ids: List[str], properties: Dict[str, Any]) -> Dict[str,int]:
        """Create task"""

        data = { 
            "taskIds": task_ids,
            **properties
        }
        
        created = self._http_client.patch('api/task-manager/projects/{}/tasks/properties'.format(project_id), json=data)

        log_raise_for_status(created)

        return get_response_json(created)

    def _get_existing_task_names_chunk(self, batch_id : BatchId, names: List[str]) -> Dict[str, List[str]]:
        """Get existing tasks
        
            Args:
                names: List of name to check (limited to 500)
           """

        existings = self._http_client.patch('api/task-manager/batches/{}/tasks-duplicate'.format(batch_id), json={ "names" : names })

        log_raise_for_status(existings)

        return get_response_json(existings)

    
    def _get_tasks_page(self, filters: TaskFilters) -> Paginated :
        """Get tasks page"""
        
        tasks = self._http_client.get('api/task-manager/tasks', params=task_filters_mapper(filters))
        
        log_raise_for_status(tasks)
        
        return get_response_json(tasks)
    


from typing import Dict, Generator
import time
from isahitlab.api.base import BaseApi
from isahitlab.domain.project import ProjectFilters
from ..helpers import get_response_json, log_raise_for_status, Paginated
from .mappers import project_filters_mapper

class ProjectApi(BaseApi):
    """Project API Calls"""

    def get_all_projects(self, filters: ProjectFilters) -> Generator[tuple[Dict, int, int], None, None] :
        """Get tasks"""

        page = 1
        limit = 10

        filters.pagination = True
        filters.limit = limit

        while True:
            filters.page = page

            paginated: Paginated = self._get_project_page(filters)
            docs = paginated['docs']
            page = paginated['nextPage']

            # Count already loaded
            totalDocs = paginated['totalDocs']
            loaded = (paginated['page'] - 1) * paginated['limit'] + len(docs)

            yield (docs, loaded, totalDocs)

            if not paginated['hasNextPage']:
                break

            time.sleep(0.5)



    
    def _get_project_page(self, filters: ProjectFilters) -> Paginated :
        """Get projects page"""
        
        tasks = self._http_client.get('api/project-manager/projects', params=project_filters_mapper(filters))
        
        log_raise_for_status(tasks)
        
        return get_response_json(tasks)


import base64
import logging
import pathlib
from datetime import datetime
from typing import Dict, List, Tuple

from isahitlab.helpers.labels import extract_labels_map_by_id

from .base import BaseFormatter

try:
    import cv2  # type: ignore
    import numpy as np
except ModuleNotFoundError as err:
    raise ModuleNotFoundError(
        "OpenCV and Numpy are required to use the mask export. To install them,"
        " run: pip install isahitlab[image]."
    ) from err


logger = logging.getLogger(__name__)

class LabToMaskFormatter(BaseFormatter):
    """Formatter to export segmentation masks 
    
    **Output folder**

    - output/
        - image1.jpg.png
        - image2.png
    

    """

    output_folder: str = None
    replace_extension: bool = False
    b_channel_mapping: Dict[int,Tuple[int,int,int]] = None

    def __init__(self, project_configuration : Dict, options : Dict = {}) -> None :
        
        if options.get("in_memory", None):
            raise Exception('"in_memory" not supported')
        
        self._init_configuration(project_configuration, options, {
            "compatible_types" : [
                "form.tool-iat-segmentation", 
                "iat-segmentation"
            ]
        })

        # Check if is sequence
        if self._project_configuration.get("metadata", {}).get("toolOptions", {}).get("sequence", False):
            raise Exception('LabToMaskFormatter does not support sequences')
        
        self.replace_extension = self._options.get('replace_extension' , False)

        if self._options.get("semantic_mapping", None):
            self.b_channel_mapping = extract_labels_map_by_id(self._project_configuration, field="bChannel") 
    
    ### Transform

    def format_tasks(self, tasks: List[Dict]) -> List[Dict]:
        formatted_tasks = []

        for task in tasks:
            formatted_task = self._format_task(task)
            formatted_tasks.append(formatted_task)

        return formatted_tasks
    
    ### Export

    def _init_export(self):
        now_str = datetime.now().strftime("%Y-%m-%d_%H:%M:%S")

        default_folder = '.'
        extra_filename_part = "_{}".format(self._options.get("extra_filename_part")) if "extra_filename_part" in self._options else ""
        default_filename = f'lab_mask{extra_filename_part}_{now_str}'
        output_folder = self._options.get("output_folder", None) or default_folder
        outputfile_path = "{}/{}".format(output_folder, self._options.get("output_filename", None) or default_filename)

        self.output_folder = outputfile_path
        path_folder = pathlib.Path(self.output_folder)
        path_folder.mkdir(parents=True, exist_ok=False)

    def append_task_to_export(self, task: Dict):
        if self.output_folder == None:
            self._init_export()

        mask = self._format_task(task)
        task_name = task["name"]
        if mask:
            self._write_mask(task_name, mask)

    def complete_export(self):
        if self.output_folder == None:
            logger.warning('Nothing to export')
            return
                
        print("Exported to \"{}\"".format(str(pathlib.Path(self.output_folder).absolute())))

    def _format_task(self, task : Dict) -> Dict:
        task = self._normalize_task(task)
        mask = task.get("data", {}).get("body", {}).get("mask", None)
        return mask

    def _write_mask(self, task_name: str, mask : str) -> None:
        if self.replace_extension and task_name.rfind('.') > -1:
            task_name = task_name[0:task_name.rfind('.')]

        file_path = pathlib.Path(self.output_folder) / f"{task_name}.png"
        file_path.parent.mkdir(parents=True, exist_ok=True)
        img = self._data_uri_to_cv2_img(mask)

        if self._options.get("semantic_mapping", None):
            img = self._transform_mask(img)

        cv2.imwrite(str(file_path), img)
    
    def _transform_mask(self, img):
        semantic_mapping = self._options.get("semantic_mapping", None)

        imgRGB = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        for m in semantic_mapping:
            b_channel = self.b_channel_mapping.get(m, None)
            if not b_channel:
                continue
            lower = np.array([1, 0, b_channel])
            upper = np.array([255, 255, b_channel])
            to = semantic_mapping[m]
            mask = cv2.inRange(imgRGB, lower, upper)
            imgRGB[mask>0] = to
        
        imgBGR = cv2.cvtColor(imgRGB, cv2.COLOR_RGB2BGR)

        return imgBGR

    def _data_uri_to_cv2_img(self, uri):
        encoded_data = uri.split(',')[1]
        decoded_data = base64.b64decode(encoded_data)
        nparr = np.fromstring(decoded_data, np.uint8)
        img = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
        return img
    

    def prepare_import(self, tasks: List[Dict]):
        raise NotImplementedError


KILI_SCHEMA = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "type": "object",
    "properties": {
        "OBJECT_DETECTION_JOB": {
            "type": "object",
            "properties": { 
                "annotations" : {
                    "type" : "array",
                    "items" : {
                        "type" : "object",
                        "properties" : {
                            "boundingPoly": {
                                "type" : "array",
                                "minContains" : 1,
                                "items" : {
                                    "type" : "object",
                                    "properties" : {
                                        "normalizedVertices" : {
                                            "type" : "array",
                                            "items" : {
                                                "type" : "object",
                                                "properties" : {
                                                    "x" : { 
                                                        "type": "number", 
                                                        "minimum": 0,
                                                        "exclusiveMaximum": 1 
                                                    },
                                                    "y" : { 
                                                        "type": "number", 
                                                        "minimum": 0,
                                                        "exclusiveMaximum": 1 
                                                    }
                                                },
                                                "required" : ["x", "y"]
                                            }
                                        },
                                        "vertices": {
                                            "type" : "array",
                                            "items" : {
                                                "type" : "object",
                                                "properties" : {
                                                    "x" : { "type": "number" },
                                                    "y" : { "type": "number" }
                                                },
                                                "required" : ["x", "y"]
                                            }
                                        }
                                    },
                                    "anyOf": [
                                        { "required" : ["normalizedVertices"] },
                                        { "required" : ["vertices"] },
                                    ]
                                }
                            },
                            "categories": {
                                "type" : "array",
                                "items" : {
                                    "type" : "object",
                                    "properties" : {
                                        "name" : { "type": "string" }
                                    }
                                }
                            }
                        },
                        "required" : ["boundingPoly"]
                    }
                }
            },
            "required": ["annotations"]
        },
        "resolution": {
          "width": { "type" : "number" },
          "height": { "type" : "number" }
        },
    },
    "required": ["OBJECT_DETECTION_JOB"],
    "$defs": {
        "annotation": {
            "type": "object",
            "properties": {
                "uid": {"type": "string"},
                "color": {"type": "string"},
                "text": {"type": "string"},
                "polygons": { "$ref": "#/$defs/polygons" },
                "labels" : { "$ref" : "#/$defs/labels" }
            },
            "required": ["polygons"],
        },
        "vertice": {
            "type": "number",
            "minimum": 0,
            "exclusiveMaximum": 1
        },
        "vertices": { 
            "type": "array", 
            "items": { "$ref" : "#/$defs/vertice" }
        },
        "polygon_type": {
            "const" : "polygon"
        },
        "polygons": {
                    "type": "array",
                    "items":  {
                        "type": "object",
                        "properties": {
                            "id":  {"type": "string"},
                            "color":  {"type": "string"},
                            "geometry":  {
                                "type": "object",
                                "properties":{
                                    "vertices" : { "$ref" : "#/$defs/vertices"
                                    },
                                    "type" : { "$ref" : "#/$defs/polygon_type" }
                                },
                                "required": ["vertices", "type"],
                            },
                            "groupId" : { "type" : "string" },
                            "text" : { "type" : "string" },
                        },
                        "required": ["geometry"],
                    }
        },
        "labels": {
            "type": "object",
            "patternProperties": {
                "^[0-9a-zA-Z-_ ]+$": {
                    "type" : "object",
                    "properties":    {
                        "labels" : {
                            "type" : "array",
                            "items" : {
                                "type" : "object",
                                "properties": {
                                    "id" : { "type" : "string"},
                                    "name" : { "type" : "string"},
                                    "displayName" : {  "oneOf": [{"type" : "string"}, { "type" : "null"}]},
                                    "uid" : { "type" : "string"},
                                },
                                "required": ["id", "name"]
                            }
                        },
                        "children" : { "$ref" : "#/$defs/labels" }
                    },
                    "additionalProperties": False,
                }
            },
            "additionalProperties": False,
        }
    },
}


from typing import Dict, Optional

from isahitlab.actions.base import BaseAction
from isahitlab.domain.batch import BatchPayload
from isahitlab.operations.batch.create_batch import CreateBatchOperation
from typeguard import typechecked


class BatchActions(BaseAction):
    """Batches actions"""

    @typechecked
    def create_batch(
        self,
        project_id: str,
        name: str,
        disable_progress_bar: Optional[bool] = False
    ) -> Dict:
        """ Create a batch in a project

        Args:
            project_id: ID of the project
            name: Name of the batch
            disable_progress_bar: Disable the progress bar display

        """

        batch = BatchPayload(name=name)

        return CreateBatchOperation(self.http_client).run(project_id=project_id, batch=batch, disable_progress_bar=disable_progress_bar)



"""HTTP client."""

import datetime
import json
import logging
from urllib.parse import urljoin

import requests
from isahitlab.core.constants import AUTH_REFRESHING_SPAN
from isahitlab.core.http.auth import Credentials
from isahitlab.exceptions import ApiBadRequest, AuthenticationFailed
import time


class HttpClient:
    """HTTP client.
    """

    def __init__(self, endpoint: str, credentials: Credentials, auth_route = "api/user-manager/auth/login", defer_auth: bool = True, retry_max_attempts = 0) -> None:
        """Initialize the HTTP client."""

        logging.basicConfig()
        self.logger = logging.getLogger('isahitlab.http')
        self.logger.setLevel(logging.DEBUG)

        self._isahit_lab_endpoint = endpoint

        self._retry_max_attempts = retry_max_attempts

        self._http_client = requests.Session()
        self._http_client_with_auth = requests.Session()

        self._credentials = credentials

        self._auth_expires_at = None

        self._auth_route = auth_route

        if not defer_auth:
            self._auth()
        # self._auth_headers = {"Authorization": f"X-API-Key: {api_key}"}
        # self._http_client_with_auth.headers.update(self._auth_headers)

    def _auth(self):
        """Get token from credentials"""

        res = self.post(urljoin(self._isahit_lab_endpoint, self._auth_route), data=self._credentials)

        if res.status_code == 401:
            raise AuthenticationFailed(self._credentials['access_id'], self._credentials["secret_key"])
        elif res.status_code == 400:
            raise ApiBadRequest(res.content)
        elif res.status_code != 201:
            raise Exception(res.content)
        
        token = json.loads(res.content)

        access_token = token['access_token']
        expires_in = token['expires_in'] #timeout in seconds

        expires_at = datetime.datetime.now() + datetime.timedelta(seconds=expires_in)
        
        self._auth_expires_at = expires_at
        self._auth_headers = {"Authorization": f"Bearer {access_token}"}
        

    def _should_authenticate(self):
        """Check for token expiration"""
        
        if self._auth_expires_at and (datetime.datetime.now() +  datetime.timedelta(seconds=AUTH_REFRESHING_SPAN) < self._auth_expires_at):
            return False

        return True

    def _send_request(self, method: str, url: str, **kwargs) -> requests.Response:
        retry_wait_s = 5

        retry_count = 0

        """Send a request to the given URL."""
        if not url.startswith('http'):
            # Get or refresh token
            if self._should_authenticate():
                self._auth()

            http_client = self._http_client_with_auth
            
            # Add authorization headers
            if kwargs.get("headers") is None:
                kwargs["headers"] = {}
            if "Authorization" not in kwargs["headers"]:
                kwargs["headers"].update(self._auth_headers)

            url = urljoin(self._isahit_lab_endpoint, url)
        else:
            http_client = self._http_client
        
        while(True):
            res = http_client.request(method, url, **kwargs)
            try:
                res.raise_for_status()
                break
            except requests.exceptions.HTTPError as err:
                if err.response.status_code >= 500 and retry_count < self._retry_max_attempts:
                    retry_count += 1
                    self.logger.warning(f"Retrying in {retry_wait_s} seconds...")
                    time.sleep(retry_wait_s)
                else:
                    break
                    

        return res

    def get(self, url: str, **kwargs) -> requests.Response:
        """Send a GET request to the given URL."""
        return self._send_request("GET", url, **kwargs)

    def post(self, url: str, **kwargs) -> requests.Response:
        """Send a POST request to the given URL."""
        return self._send_request("POST", url, **kwargs)

    def head(self, url: str, **kwargs) -> requests.Response:
        """Send a HEAD request to the given URL."""
        return self._send_request("HEAD", url, **kwargs)

    def put(self, url: str, **kwargs) -> requests.Response:
        """Send a PUT request to the given URL."""
        return self._send_request("PUT", url, **kwargs)

    def patch(self, url: str, **kwargs) -> requests.Response:
        """Send a PATCH request to the given URL."""
        return self._send_request("PATCH", url, **kwargs)

    def options(self, url: str, **kwargs) -> requests.Response:
        """Send a OPTIONS request to the given URL."""
        return self._send_request("OPTIONS", url, **kwargs)



from typing import Any, Dict, List, NewType

InputsLayout = NewType("InputsLayout", List[List[List[Dict[str,Any]]]])

IAT_INPUT_TYPES = [
    "tool-iat-rectangle", 
    "tool-iat-polygon", 
    "tool-iat-polyline",
    "tool-iat-graph",
    "tool-iat-segmentation", 
]

def extract_all_inputs_from_layout(layout: InputsLayout) -> List[Dict[str, Any]]:
    """Extract the inputs from the layouts"""
    inputs: List[Dict[str,Dict]] = []

    for row in layout:
        for col in row:
            for input in col:
                inputs.append(input)

    return inputs

def extract_not_static_inputs_from_layout(layout: InputsLayout) -> List[Dict[str, Any]]:
    """Extract from the layouts only the inputs that can be set for task initialization"""
    inputs: List[Dict[str,Dict]] = []

    for row in layout:
        for col in row:
            for input in col:
                if input['submit'] != False or not input.get('static', False):
                    inputs.append(input)

    return inputs

def extract_inputs_by_types_from_layout(layout: InputsLayout, types: List[str]) -> List[Dict[str, Any]]:
    """Extract from the layouts only the inputs with the provided type"""
    inputs: List[Dict[str,Dict]] = []

    for row in layout:
        for col in row:
            for input in col:
                if input['type'] in types:
                    inputs.append(input)

    return inputs

def extract_inputs_iat_from_layout(layout: InputsLayout):
    """Extract from the layouts only the inputs with the provided type"""
    inputs: List[Dict[str,Dict]] = []

    for row in layout:
        for col in row:
            for input in col:
                if input['type'][0:8] == "tool-iat":
                    inputs.append(input)

    return inputs

def extract_inputs_by_id_from_layout(layout: InputsLayout, id: str) -> List[Dict[str, Any]]:
    """Extract from the layouts only the input with the id"""
    inputs: List[Dict[str,Dict]] = []

    for row in layout:
        for col in row:
            for input in col:
                if input['id'] == id:
                    inputs.append(input)

    return inputs


import json
import time
from typing import Dict, Generator, List, Optional, Union

from isahitlab.api.base import BaseApi
from isahitlab.domain.batch import BatchId
from isahitlab.domain.workflow import WebhookPayload, WorkflowId

from ..helpers import get_response_json, log_raise_for_status


class WorkflowApi(BaseApi):
    """Task API Calls"""

    def get_workflow_by_id(self, workflow_id: WorkflowId) -> Dict :
        """Get webhook of workflow"""

        result = self._http_client.get(f'api/workflow-manager/workflows/{workflow_id}')
        
        log_raise_for_status(result)

        return get_response_json(result)
    

    def setup_webhook_by_workflow_id(self, workflow_id : WorkflowId, webhook : WebhookPayload):
        """Setup webhook of workflow"""

        data = {
            "url" : webhook.url,
            "header": webhook.header
        }

        result = self._http_client.patch('api/workflow-manager/workflows/setup-webhook', params={ "workflowId" : workflow_id }, json=data)
        
        log_raise_for_status(result)
    
    def setup_webhook_by_batch_id(self, batch_id : BatchId, webhook : WebhookPayload):
        """Setup webhook of batch"""

        data = {
            "url" : webhook.url,
            "header": webhook.header
        }

        result = self._http_client.patch('api/workflow-manager/workflows/setup-webhook', params={ "batchId" : batch_id }, json=data)
        
        log_raise_for_status(result)

    


from tqdm import tqdm
from typing import Optional, Dict
from isahitlab.operations.base import BaseAction
from isahitlab.api.project_configuration.api import ProjectConfigurationApi

from typeguard import typechecked

class GetProjectConfigurationOperation(BaseAction):
    """Project configuration actions"""

    @typechecked
    def run(
        self,
        project_id: str,
        disable_progress_bar: Optional[bool] = False
    ) -> Dict:
        """ Get the configuration of a project
        
        Args:
            project_id: ID of the project
            disable_progress_bar: Disable the progress bar display
        """

        project_configuration = None

        project_configuration_api = ProjectConfigurationApi(self._http_client)
        with tqdm(total=1,  disable=disable_progress_bar, desc="Loading project configuration... ") as loader:
            project_configuration = project_configuration_api.get_project_configuration(project_id)
            loader.update(1)

        return project_configuration
    


import warnings
from typing import Any, Dict, Generator, Iterable, List, Optional, Union, cast

from isahitlab.actions.base import BaseAction
from isahitlab.domain.batch import BatchId
from isahitlab.domain.export import ExportFormat
from isahitlab.domain.project import ProjectFilters, ProjectId, ProjectType
from isahitlab.domain.task import (TaskCompatibilityMode, TaskFilters, TaskId,
                                   TaskOptionalFields, TaskPayload, TaskStatus)
from isahitlab.operations.project.get_projects import GetProjectsOperation
from isahitlab.operations.task.create_tasks import CreateTaskOperation
from isahitlab.operations.task.export import ExportOperation
from isahitlab.operations.task.get_tasks import GetTasksOperation
from isahitlab.operations.task.update_properties import UpdatePropertiesOperation
from isahitlab.helpers.list import remove_duplicates
from typeguard import typechecked


class TaskActions(BaseAction):
    """Tasks actions"""

    @typechecked
    def tasks(
        self,
        project_id: ProjectId,
        batch_id_in: Optional[List[str]] = None,
        status_in: Optional[List[str]] = None,
        task_id_in: Optional[List[str]] = None,
        name_in: Optional[List[str]] = None,
        name_like: Optional[str] = None,
        created_at_gt: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lt: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        updated_at_gt: Optional[str] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lt: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
        optional_fields: Optional[List[TaskOptionalFields]] = ["data"],
        disable_progress_bar: Optional[bool] = False,
        compatibility_mode: Optional[TaskCompatibilityMode] = None,
        raw_data: Optional[bool] = False,
        iterate: Optional[bool] = False
    ) -> Union[Generator[Iterable[Dict], None, None], Iterable[Dict]]:
        """ Get an task list

        !!! warning "No data by default"
            Add `"data"` in the `optional_fields` if you want to get your task data

        Args:
            project_id: ID of the project
            batch_id_in: A list of batch ids to filter
            status_in: Only in those statuses.
                Possible choices: `pending`, `complete`, `to-review`, `reviewed`, `configuring`.
            task_id_in: Only task whose ID is in this list,
            name_in: Only task whose name is in this list,
            name_like: Only task whose name mathes this regex,
            created_at_gt: Only tasks created after this date (Ex.: "2022-09-19 08:30:00")
            created_at_gte: Only tasks created at or after this date (Ex.: "2022-09-19 08:30:00")
            created_at_lt: Only tasks created before this date (Ex.: "2022-09-19 08:30:00")
            created_at_lte: Only tasks created at or before this date (Ex.: "2022-09-19 08:30:00")
            updated_at_gt: Only tasks updated after this date (Ex.: "2022-09-19 08:30:00")
            updated_at_gte: Only tasks updated at or after this date (Ex.: "2022-09-19 08:30:00")
            updated_at_lt: Only tasks updated before this date (Ex.: "2022-09-19 08:30:00")
            updated_at_lte: Only tasks updated at or before this date (Ex.: "2022-09-19 08:30:00")
            optional_fields: retreive those additional information
                Possible choices: `data`, `jobs`, `metrics`, `data.mask`(for segmentation project).
            disable_progress_bar: Disable the progress bar display
            compatibility_mode: Format the output for specific use cases
                Possible choices: `kili` -> format the ouput to look like kili.assets() results 
            raw_data: Get raw data and metadata object instead of merged data
            iterate: Return a generator

        Returns:
            List (or Generator) of tasks representations
        """

        filters = TaskFilters(
            project_id=project_id,
            batch_id_in=cast(
                List[BatchId], batch_id_in) if batch_id_in else None,
            status_in=cast(List[TaskStatus], status_in) if status_in else None,
            task_id_in=cast(List[TaskId], task_id_in) if task_id_in else None,
            name_in=cast(List[str], name_in) if name_in else None,
            name_like=name_like if name_like else None,
            created_at_gt=cast(
                List[str], created_at_gt) if created_at_gt else None,
            created_at_gte=cast(
                List[str], created_at_gte) if created_at_gte else None,
            created_at_lt=cast(
                List[str], created_at_lt) if created_at_lt else None,
            created_at_lte=cast(
                List[str], created_at_lte) if created_at_lte else None,
            updated_at_gt=cast(
                List[str], updated_at_gt) if updated_at_gt else None,
            updated_at_gte=cast(
                List[str], updated_at_gte) if updated_at_gte else None,
            updated_at_lt=cast(
                List[str], updated_at_lt) if updated_at_lt else None,
            updated_at_lte=cast(
                List[str], updated_at_lte) if updated_at_lte else None,
            optional_fields=cast(
                List[TaskOptionalFields], optional_fields) if optional_fields else None
        )

        operation = GetTasksOperation(self.http_client)

        operation_gen = operation.run(project_id=project_id, filters=filters,
                                      disable_progress_bar=disable_progress_bar or iterate, compatibility_mode=compatibility_mode, raw_data=raw_data)

        if iterate:
            return operation_gen
        else:
            return list(operation_gen)

    @typechecked
    def create_tasks(
        self,
        project_id: str,
        batch_id: str,
        tasks: List[Dict],
        disable_progress_bar: Optional[bool] = False,
        disable_data_check: Optional[bool] = False,
        disable_unicity_check: Optional[bool] = False,
        raise_if_existing: Optional[bool] = True,
        compatibility_mode: Optional[TaskCompatibilityMode] = None
    ) -> Iterable[Dict]:
        """ Create tasks in a batch

        Args:
            project_id: ID of the project
            batch_id: ID of the batch project
            tasks: List of the tasks to create. See TaskPayload(name: str, resource: str, data: Dict[str, Any], properties: Optional[Dict[str, Any]])
            disable_progress_bar: Disable the progress bar display
            disable_data_check: Set this option to False to ignore the data format validation
            disable_unicity_check: Set this option to True to ignore duplicate check and insert it anyway
            raise_if_existing: Set this option to False to skip task duplicates and only show a warning
            compatibility_mode: Format the output for specific use cases
                Possible choices: `kili` -> format the ouput to look like kili.assets() results 

        """

        projects = list(GetProjectsOperation(self.http_client).run(filters=ProjectFilters(
            id_in=[project_id]), disable_progress_bar=disable_progress_bar))
        project = projects[0] if len(projects) > 0 else None

        if not project:
            warnings.warn('Project not found')
            return

        project_type: ProjectType = project['type']

        operation = CreateTaskOperation(self.http_client)
        task_payloads = [*map(lambda t: TaskPayload(
            name=t['name'],
            properties=t.get('properties', None),
            resources=t['resources'] if project_type != "form" else None,
            data=t['data'] if 'data' in t else None
        ), tasks)]

        tasks = operation.run(project_id=project_id,
                              batch_id=batch_id,
                              tasks=task_payloads,
                              disable_progress_bar=disable_progress_bar,
                              disable_unicity_check=disable_unicity_check,
                              raise_if_existing=raise_if_existing,
                              disable_data_check=disable_data_check,
                              compatibility_mode=compatibility_mode
                              )

        return tasks

    @typechecked
    def update_properties_of_tasks(
        self,
        project_id: str,
        task_id_in: List[str],
        properties: Dict[str, Any],
        disable_progress_bar: Optional[bool] = False
    ) -> int:
        """ Update properties of tasks

        Args:
            project_id: ID of the project
            task_id_in: List of the IDs of the tasks to update
            properties: Dict of properties
                Possible keys : `score`, `custom`
            disable_progress_bar: Disable the progress bar display

        !!! info "Custom properties"
            Some properties like `score` are known by the system. 
            To add custom properties, pass a dictionnary of properties in `properties.custom` (see example below).
    

        !!! Example
            ```python
            from isahitlab.client import IsahitLab

            lab = IsahitLab()

            lab.update_properties_of_tasks(
                        project_id='<project_id>', 
                        task_id_in=['<task_id_1>', '<task_id_2>'],
                        properties={
                            "score": 5,
                            "custom": {
                                "my-property" : "my_value"
                            }
                        }
            )

            ```
        """

        task_id_in = remove_duplicates(task_id_in)

        for property in properties.keys():
            if property not in ["score", "custom"]:
                raise ValueError(f"Unknow property {property}")
            
        return UpdatePropertiesOperation(self.http_client).run(
            project_id=project_id, task_id_in=task_id_in, properties=properties, disable_progress_bar=disable_progress_bar)
       

    @typechecked
    def export_tasks(
        self,
        project_id: ProjectId,
        format: ExportFormat,
        output_folder: Optional[str] = None,
        output_filename: Optional[str] = None,
        in_memory: Optional[bool] = False,
        options: Optional[Dict] = {},
        batch_id_in: Optional[List[str]] = None,
        status_in: Optional[List[str]] = None,
        task_id_in: Optional[List[str]] = None,
        name_in: Optional[List[str]] = None,
        name_like: Optional[str] = None,
        created_at_gt: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lt: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        updated_at_gt: Optional[str] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lt: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
        disable_progress_bar: Optional[bool] = False
    ) -> Any:
        """ Get an task list

        Args:
            project_id: ID of the project
            format: Export format
                Possible choices: `lab`, `kili`, `yolo`, `mask`
            output_folder: Path to the folder where the export will be saved
            output_filename: name of the export
            in_memory: Return the result instead of saving it on file system
            options: Specific options according to the format (see documentation)
            batch_id_in: A list of batch ids to filter
            status_in: Only in those statuses.
                Possible choices: `pending`, `complete`, `to-review`, `reviewed`, `configuring`.
            task_id_in: Only task whose ID is in this list,
            name_in: Only task whose name is in this list,
            name_like: Only task whose name mathes this regex,
            created_at_gt: Only tasks created after this date (Ex.: "2022-09-19 08:30:00")
            created_at_gte: Only tasks created at or after this date (Ex.: "2022-09-19 08:30:00")
            created_at_lt: Only tasks created before this date (Ex.: "2022-09-19 08:30:00")
            created_at_lte: Only tasks created at or before this date (Ex.: "2022-09-19 08:30:00")
            updated_at_gt: Only tasks updated after this date (Ex.: "2022-09-19 08:30:00")
            updated_at_gte: Only tasks updated at or after this date (Ex.: "2022-09-19 08:30:00")
            updated_at_lt: Only tasks updated before this date (Ex.: "2022-09-19 08:30:00")
            updated_at_lte: Only tasks updated at or before this date (Ex.: "2022-09-19 08:30:00")
            disable_progress_bar: Disable the progress bar display

        """

        filters = TaskFilters(
            project_id=project_id,
            batch_id_in=cast(
                List[BatchId], batch_id_in) if batch_id_in else None,
            status_in=cast(List[TaskStatus], status_in) if status_in else None,
            task_id_in=cast(List[TaskId], task_id_in) if task_id_in else None,
            name_in=cast(List[str], name_in) if name_in else None,
            name_like=name_like if name_like else None,
            created_at_gt=cast(
                List[str], created_at_gt) if created_at_gt else None,
            created_at_gte=cast(
                List[str], created_at_gte) if created_at_gte else None,
            created_at_lt=cast(
                List[str], created_at_lt) if created_at_lt else None,
            created_at_lte=cast(
                List[str], created_at_lte) if created_at_lte else None,
            updated_at_gt=cast(
                List[str], updated_at_gt) if updated_at_gt else None,
            updated_at_gte=cast(
                List[str], updated_at_gte) if updated_at_gte else None,
            updated_at_lt=cast(
                List[str], updated_at_lt) if updated_at_lt else None,
            updated_at_lte=cast(
                List[str], updated_at_lte) if updated_at_lte else None,
            optional_fields=["data", "data.mask", "jobs", "metrics"]
        )

        operation = ExportOperation(self.http_client)

        return operation.run(project_id=project_id, filters=filters, format=format, disable_progress_bar=disable_progress_bar, options={
            **options,
            "output_folder": output_folder,
            "output_filename": output_filename,
            "in_memory": in_memory
        })



from typing import Dict, Optional


def merge_data_and_metadata(data_body: Optional[Dict], metadata_body : Optional[Dict]):
    data_body = data_body or {}
    metadata_body = metadata_body or {}

    keys = set([*data_body.keys(),*metadata_body.keys()])

    merged = {}

    for key in keys:
        value_in_data = data_body.get(key, None)
        value_in_metadata = metadata_body.get(key, None)
    
        if value_in_data and value_in_metadata and isinstance(value_in_data, dict) and isinstance(value_in_metadata, dict):
          merged[key] = {
            **value_in_metadata,
            **value_in_data
          }
        elif value_in_data != None:
          merged[key] = value_in_data
        elif value_in_metadata != None:
          merged[key] = value_in_metadata

    return merged


import time
from typing import Dict, Generator

from isahitlab.api.base import BaseApi
from isahitlab.domain.batch import BatchFilters, BatchId, BatchPayload
from isahitlab.domain.project import ProjectId

from ..helpers import Paginated, get_response_json, log_raise_for_status
from .mappers import batch_filters_mapper


class BatchApi(BaseApi):
    """Batch API Calls"""

    def get_batch_by_id(self, batch_id: BatchId) -> Dict:
        """Get batch"""

        project_configuration = self._http_client.get(
            'api/task-manager/batches/{}'.format(batch_id))

        log_raise_for_status(project_configuration)

        return get_response_json(project_configuration)

    def get_all_batches(self, filters: BatchFilters) -> Generator[tuple[Dict, int, int], None, None]:
        """Get batches"""

        page = 1
        limit = 10

        filters.pagination = True
        filters.limit = limit

        while True:
            filters.page = page

            paginated: Paginated = self._get_batches_page(filters)
            docs = paginated['docs']
            page = paginated['nextPage']

            # Count already loaded
            totalDocs = paginated['totalDocs']
            loaded = (paginated['page'] - 1) * paginated['limit'] + len(docs)

            yield (docs, loaded, totalDocs)

            if not paginated['hasNextPage']:
                break

            time.sleep(0.5)

    def _get_batches_page(self, filters: BatchFilters) -> Paginated:
        """Get batches page"""

        tasks = self._http_client.get(
            'api/task-manager/batches', params=batch_filters_mapper(filters))

        log_raise_for_status(tasks)

        return get_response_json(tasks)

    def get_batches_by_project_id(self, project_id: ProjectId) -> Dict:
        """Get batches of project"""

        project_configuration = self._http_client.get('api/task-manager/batches', params={
            "projectId": project_id
        })

        log_raise_for_status(project_configuration)

        return get_response_json(project_configuration)


    def create_batch(self, project_id: ProjectId, batch: BatchPayload):
        """Create batch in a project"""

        data = {
            "name": batch.name,
            "projectId": project_id
        }

        result = self._http_client.post('api/task-manager/batches', json=data)

        log_raise_for_status(result)

        return get_response_json(result)



import logging
from typing import List, Optional, Dict, Any

from isahitlab.api.task.api import TaskApi
from isahitlab.domain.project import ProjectId
from isahitlab.domain.task import TaskId
from isahitlab.operations.base import BaseAction
from isahitlab.helpers.list import divide_chunks
from tqdm import tqdm
from typeguard import typechecked

logger = logging.getLogger('isahitlab.update_properties')


class UpdatePropertiesOperation(BaseAction):
    """Update properties of tasks
    """

    _task_api : TaskApi

    @typechecked
    def run(
        self,
        project_id: ProjectId,
        task_id_in: List[TaskId],
        properties: Dict[str, Any],
        disable_progress_bar: Optional[bool] = False,
    ) -> int:
        """ Create tasks in a batch
        
        Args:
            project_id: ID of the project
            task_id_in: List of the IDs of the tasks to update
            properties; Dict of properties
                Possible keys : `score`
            disable_progress_bar: Disable the progress bar display

        """
        self._task_api = TaskApi(self._http_client)
        
        chunks = divide_chunks(task_id_in, 200)
        updated = 0
        
        with tqdm(total=len(task_id_in),  disable=disable_progress_bar, desc="Updating tasks properties... ") as loader:
            for task_ids in chunks:
                result = self._task_api.update_properties(project_id=project_id, task_ids=task_ids, properties=properties)
                updated += result["updated"]
                loader.update(len(task_ids))
            

        
        return updated
    


from abc import ABC
from isahitlab.core.http.http_client import HttpClient


class BaseAction(ABC):
    """Base class for all actions.
    It is used to share the HttpClient between all actions classes.
    It is not meant to be used and instantiated directly.
    """
    
    http_client: HttpClient # instantianted in the isahitlab client child class


import base64
from math import floor
from typing import Dict

try:
    import cv2  # type: ignore
    import numpy as np
except ModuleNotFoundError as err:
    raise ModuleNotFoundError(
        "OpenCV and Numpy are required to use the mask helpers. To install them,"
        " run: pip install isahitlab[image]."
    ) from err

from .annotation import clean_up
from .color import to_bgr
from .labels import extract_main_labels
from .points import denormalize_points, vertices_to_points, points_to_vertices, bbox_to_corners


def polygon_data_to_segmentation_data(data: Dict, labels_mapping_by_id : Dict[str, Dict], image_size : Dict[str, int]) -> Dict :
    
    data = clean_up(data)

    segmentation_task = {
        **data,
        "annotations": [],
        "image" : image_size 
    }

    mask = np.zeros((image_size["height"], image_size["width"], 3))

    annotations = data.get('annotations', [])

    instance_id = 2

    for annotation in annotations:
        polygons = annotation.get("polygons", [])
        if len(polygons) == 0:
            continue

        # Resolve B Channel
        labels = annotation["labels"] if "labels" in annotation else {}
        main_labels = extract_main_labels(labels)
    
        matching_label_in_target = labels_mapping_by_id.get(main_labels[0]['id'], None) if len(main_labels) > 0 else None
        
        r = instance_id % 256
        g = floor(instance_id / 256)
        if matching_label_in_target:
            b = matching_label_in_target["bChannel"]
        else:
            b = 1
        rgb = (r, g, b)

        # Handle polygons
        cv_polygons = []
        for polygon in polygons:
            normalized_vertices =  polygon["geometry"]["vertices"]
            geometry_type = polygon["geometry"]["type"]

            if geometry_type == "rectangle":
                normalized_vertices = bbox_to_corners(normalized_vertices)

            polygon["color"] = matching_label_in_target.get("color", polygon["color"]) if matching_label_in_target else polygon["color"]
            denormalized_points = denormalize_points(vertices_to_points(normalized_vertices), image_size, 0)
            
            # Approximation are denormalized in the lab
            polygon["geometry"]["vertices"] = points_to_vertices(denormalized_points)
            polygon["geometry"]["type"] = "polygon"

            cv_polygon = np.array([(point['x'], point['y'])
                               for point in denormalized_points], dtype=np.int32)
            cv_polygons.append(cv_polygon)
            
        cv2.fillPoly(mask, cv_polygons, to_bgr(rgb))

        # Create annotation
        segmentation_task["annotations"].append({
            **annotation,
            "color" : matching_label_in_target.get("color", annotation["color"]) if matching_label_in_target else annotation["color"],
            "polygons": list(rgb),
            "approximation" : {
                "polygons": polygons,
                "minus" : []
            }
        })

        instance_id += 1

    # Mask to base64
    retval, buffer = cv2.imencode('.png', mask)
    jpg_as_text = base64.b64encode(buffer)
    base64mask = 'data:image/png;base64,' + jpg_as_text.decode()
    segmentation_task["mask"] = base64mask
    return segmentation_task



from typing import Dict, Any
from isahitlab.api.base import BaseApi
from ..helpers import get_response_json, log_raise_for_status

class DataApi(BaseApi):
    """Data API Calls"""

    def import_data(self, batch_id: str, name : str, body: Dict[str, Any]) -> Dict :
        """Get files grant token"""
        
        data = self._http_client.post('api/data-manager/import-data/json', 
                                      params= {
                                        "skipTaskCreation" : "true",
                                        "grantAccess" : "true"

                                    }, 
                                    json={ 
                                        "batchId": batch_id,
                                        "name" : name,
                                        "body" : body
                                    }
        )
        
        log_raise_for_status(data)
        
        return get_response_json(data)
    